from teradataml import (
    create_context,
    get_context
)
from feast.repo_config import FeastConfigBaseModel
from typing import Dict, Optional
from pydantic import StrictStr
from feast.value_type import ValueType

import pyarrow as pa


class TeradataConfig(FeastConfigBaseModel):
    host: StrictStr
    port: int = 1025
    database: StrictStr
    user: StrictStr
    password: StrictStr
    log_mech: Optional[StrictStr] = "LDAP"


def teradata_type_to_feast_value_type(data_type):
    type_map: Dict[str, ValueType] = {
        "<class 'int'>": pa.int64(),
        "<class 'float'>": pa.float64(),
        "<class 'datetime.datetime'>": pa.timestamp('ns'),
        "<class 'str'>": pa.string()
    }

    return type_map[str(data_type)]


def get_conn(config: TeradataConfig):
    if get_context() is None:
        create_context(host=config.host,
                       username=config.user,
                       password=config.password,
                       database=config.database,
                       logmech=config.log_mech)

    return get_context()



import click
import os

from feast.file_utils import replace_str_in_file
from teradataml import (
    create_context,
    get_context
)


def bootstrap():
    # Bootstrap() will automatically be called from the init_repo() during `feast init`

    import pathlib
    from datetime import datetime, timedelta

    from feast.driver_test_data import create_driver_hourly_stats_df

    repo_path = pathlib.Path(__file__).parent.absolute()
    project_name = os.path.basename(repo_path)

    repo_path = repo_path / "feature_repo"

    end_date = datetime.now().replace(microsecond=0, second=0, minute=0)
    start_date = end_date - timedelta(days=15)

    driver_entities = [1001, 1002, 1003, 1004, 1005]
    driver_df = create_driver_hourly_stats_df(driver_entities, start_date, end_date)

    data_path = repo_path / "data"
    data_path.mkdir(exist_ok=True)
    driver_stats_path = data_path / "driver_stats.parquet"
    driver_df.to_parquet(path=str(driver_stats_path), allow_truncated_timestamps=True)

    # check if running in interactive mode or via CI
    is_interactive = os.environ.get("CI_RUNNING", "false").lower() == "false"
    if is_interactive:
        host = click.prompt("Teradata Host URL")
        teradata_user = click.prompt("Teradata User Name")
        teradata_password = click.prompt("Teradata Password", hide_input=True)
        teradata_database = click.prompt("Teradata Database Name")
        teradata_log_mech = click.prompt("Teradata Connection Mechanism")

    else:
        host = os.environ.get("CI_TD_HOST")
        teradata_user = os.environ.get("CI_TD_USER")
        teradata_password = os.environ.get("CI_TD_PASSWORD")
        teradata_database = os.environ.get("CI_TD_DATABASE")
        teradata_log_mech = os.environ.get("CI_TD_LOGMECH")

    config_file = repo_path / "feature_store.yaml"
    path = 'teradatasql://'+ teradata_user +':' + teradata_password + '@'+host + '/?database=' + teradata_database + '&LOGMECH=' + teradata_log_mech
    replace_str_in_file(config_file, "Teradata registry path", path)
    for i in range(2):
        replace_str_in_file(
            config_file, "Teradata host URL", host
        )
        replace_str_in_file(config_file, "Teradata user", teradata_user)
        replace_str_in_file(config_file, "Teradata password", teradata_password)
        replace_str_in_file(config_file, "Teradata database", teradata_database)
        replace_str_in_file(config_file, "Teradata log_mech", teradata_log_mech)

    if not is_interactive or click.confirm(
        f'Should I upload example data to Teradata (overwriting "{project_name}_feast_driver_hourly_stats" table)?',
        default=True,
    ):
        if get_context() is None:
            create_context(host=host,
                           username=teradata_user,
                           password=teradata_password,
                           database=teradata_database,
                           logmech=teradata_log_mech)
        teradata_conn = get_context().connect()

        with teradata_conn as conn:
            driver_df.to_sql(name=f"{project_name}_feast_driver_hourly_stats", con=conn, if_exists="replace",
                             index=False)



import contextlib
from dataclasses import asdict
from datetime import datetime
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterator,
    KeysView,
    List,
    Optional,
    Tuple,
    Union,
)
import pyarrow
import numpy as np
import pandas as pd
import pyarrow as pa
from jinja2 import BaseLoader, Environment
from pydantic.typing import Literal
from pytz import utc

from feast.data_source import DataSource
from feast.errors import InvalidEntityType
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores import offline_utils

from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.registry.registry import Registry
from feast_teradata.teradata_utils import (
    get_conn,
    TeradataConfig,
    teradata_type_to_feast_value_type,
)
from feast_teradata.offline.teradata_source import (
    TeradataSource,
    df_to_teradata_table
)
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.usage import log_exceptions_and_usage


class TeradataOfflineStoreConfig(TeradataConfig):
    type: Literal[
        "feast_teradata.offline.teradata.TeradataOfflineStore"
    ] = "feast_teradata.offline.teradata.TeradataOfflineStore"


class TeradataOfflineStore(OfflineStore):
    @staticmethod
    @log_exceptions_and_usage(offline_store="teradata")
    def pull_latest_from_table_or_query(
            config: RepoConfig,
            data_source: DataSource,
            join_key_columns: List[str],
            feature_name_columns: List[str],
            timestamp_field: str,
            created_timestamp_column: Optional[str],
            start_date: datetime,
            end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, TeradataOfflineStoreConfig)
        assert isinstance(data_source, TeradataSource)
        from_expression = data_source.get_table_query_string()

        partition_by_join_key_string = ", ".join(_append_alias(join_key_columns, "a"))
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                    "PARTITION BY " + partition_by_join_key_string
            )
        timestamps = [timestamp_field]
        if created_timestamp_column:
            timestamps.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(_append_alias(timestamps, "a")) + " DESC"
        a_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "a")
        )
        b_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "b")
        )

        query = f"""
            SELECT
                {b_field_string}
                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
            FROM (
                SELECT {a_field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression} a
                WHERE a."{timestamp_field}" BETWEEN '{start_date}' AND '{end_date}'
            ) b
            WHERE _feast_row = 1
            """


        return TeradataRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="teradata")
    def get_historical_features(
            config: RepoConfig,
            feature_views: List[FeatureView],
            feature_refs: List[str],
            entity_df: Union[pd.DataFrame, str],
            registry: Registry,
            project: str,
            full_feature_names: bool = False,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, TeradataOfflineStoreConfig)
        for fv in feature_views:
            assert isinstance(fv.batch_source, TeradataSource)

        entity_schema = _get_entity_schema(entity_df, config)

        entity_df_event_timestamp_col = (
            offline_utils.infer_event_timestamp_from_entity_df(entity_schema)
        )

        entity_df_event_timestamp_range = _get_entity_df_event_timestamp_range(
            entity_df,
            entity_df_event_timestamp_col,
            config,
        )

        @contextlib.contextmanager
        def query_generator() -> Iterator[str]:
            table_name = offline_utils.get_temp_entity_table_name()

            _upload_entity_df(config, entity_df, table_name)

            expected_join_keys = offline_utils.get_expected_join_keys(
                project, feature_views, registry
            )

            offline_utils.assert_expected_columns_in_entity_df(
                entity_schema, expected_join_keys, entity_df_event_timestamp_col
            )

            query_context = offline_utils.get_feature_view_query_context(
                feature_refs,
                feature_views,
                registry,
                project,
                entity_df_event_timestamp_range,
            )

            query_context_dict = [asdict(context) for context in query_context]
            # Hack for query_context.entity_selections to support uppercase in columns
            for context in query_context_dict:
                context["entity_selections"] = [
                    f'''"{entity_selection.replace(' AS ', '" AS "')}\"'''
                    for entity_selection in context["entity_selections"]
                ]

            try:
                yield build_point_in_time_query(
                    query_context_dict,
                    left_table_query_string=table_name,
                    entity_df_event_timestamp_col=entity_df_event_timestamp_col,
                    entity_df_columns=entity_schema.keys(),
                    query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
                    full_feature_names=full_feature_names,
                )
            finally:
                if table_name:
                    with get_conn(config.offline_store).connect() as conn:
                        conn.execute(f"DROP TABLE {table_name}")

        return TeradataRetrievalJob(
            query=query_generator,
            config=config,
            full_feature_names=full_feature_names,
            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
                feature_refs, project, registry
            ),
            metadata=RetrievalMetadata(
                features=feature_refs,
                keys=list(entity_schema.keys() - {entity_df_event_timestamp_col}),
                min_event_timestamp=entity_df_event_timestamp_range[0],
                max_event_timestamp=entity_df_event_timestamp_range[1],
            ),
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="teradata")
    def pull_all_from_table_or_query(
            config: RepoConfig,
            data_source: DataSource,
            join_key_columns: List[str],
            feature_name_columns: List[str],
            timestamp_field: str,
            start_date: datetime,
            end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, TeradataOfflineStoreConfig)
        assert isinstance(data_source, TeradataSource)
        from_expression = data_source.get_table_query_string()

        field_string = ", ".join(
            join_key_columns + feature_name_columns + [timestamp_field]
        )

        start_date = start_date.astimezone(tz=utc)
        end_date = end_date.astimezone(tz=utc)

        query = f"""
            SELECT {field_string}
            FROM {from_expression} AS paftoq_alias
            WHERE "{timestamp_field}" BETWEEN '{start_date}'::timestamptz AND '{end_date}'::timestamptz
        """

        return TeradataRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
        )
    @staticmethod
    def offline_write_batch(
        config: RepoConfig,
        feature_view: FeatureView,
        table: pyarrow.Table,
        progress: Optional[Callable[[int], Any]],
    ):
        assert isinstance(config.offline_store, TeradataOfflineStoreConfig)
        assert isinstance(feature_view.batch_source, TeradataSource)

        pa_schema, column_names = offline_utils.get_pyarrow_schema_from_batch_source(
            config, feature_view.batch_source
        )
        if column_names != table.column_names:
            raise ValueError(
                f"The input pyarrow table has schema {table.schema} with the incorrect columns {table.column_names}. "
                f"The schema is expected to be {pa_schema} with the columns (in this exact order) to be {column_names}."
            )

        if table.schema != pa_schema:
            table = table.cast(pa_schema)

        table_df = table.to_pandas()
        with get_conn(config.offline_store).connect() as conn:
            table_df.to_sql(name=feature_view.batch_source.name,
                            con=conn,
                            if_exists="append",
                            index=False)

class TeradataRetrievalJob(RetrievalJob):
    def __init__(
            self,
            query: Union[str, Callable[[], ContextManager[str]]],
            config: RepoConfig,
            full_feature_names: bool,
            on_demand_feature_views: Optional[List[OnDemandFeatureView]],
            metadata: Optional[RetrievalMetadata] = None,
    ):
        if not isinstance(query, str):
            self._query_generator = query
        else:

            @contextlib.contextmanager
            def query_generator() -> Iterator[str]:
                assert isinstance(query, str)
                yield query

            self._query_generator = query_generator
        self.config = config
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []
        self._metadata = metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        # We use arrow format because it gives better control of the table schema
        return self._to_arrow_internal().to_pandas()

    def to_sql(self) -> str:
        with self._query_generator() as query:
            return query

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        with self._query_generator() as query:
            with get_conn(self.config.offline_store).raw_connection().cursor() as cur:
                cur.execute(query)
                fields = [
                    (c[0], teradata_type_to_feast_value_type(c[1]))
                    for c in cur.description
                ]
                data = cur.fetchall()
                schema = pa.schema(fields)

                data_transposed: List[List[Any]] = []
                for col in range(len(fields)):
                    data_transposed.append([])
                    for row in range(len(data)):
                        data_transposed[col].append(data[row][col])
                table = pa.Table.from_arrays(
                    [pa.array(row) for row in data_transposed], schema=schema
                )

            return table

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    def persist(self, storage: SavedDatasetStorage, allow_overwrite: bool = False):
        raise NotImplementedError("Not yet implemented")


def _get_entity_df_event_timestamp_range(
        entity_df: Union[pd.DataFrame, str],
        entity_df_event_timestamp_col: str,
        config: RepoConfig,
) -> Tuple[datetime, datetime]:
    if isinstance(entity_df, pd.DataFrame):
        entity_df_event_timestamp = entity_df.loc[
                                    :, entity_df_event_timestamp_col
                                    ].infer_objects()
        if pd.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pd.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )
    elif isinstance(entity_df, str):
        # If the entity_df is a string (SQL query), determine range
        # from table
        with get_conn(config.offline_store).raw_connection().cursor() as cur:
            cur.execute(
                f"SELECT MIN({entity_df_event_timestamp_col}) AS min_ts, MAX({entity_df_event_timestamp_col}) AS max_ts FROM ({entity_df}) as tmp_alias"
            ),
            res = cur.fetchone()
        entity_df_event_timestamp_range = (res[0], res[1])
    else:
        raise InvalidEntityType(type(entity_df))

    return entity_df_event_timestamp_range


def _append_alias(field_names: List[str], alias: str) -> List[str]:
    return [f'{alias}."{field_name}"' for field_name in field_names]


def build_point_in_time_query(
        feature_view_query_contexts: List[dict],
        left_table_query_string: str,
        entity_df_event_timestamp_col: str,
        entity_df_columns: KeysView[str],
        query_template: str,
        full_feature_names: bool = False,
) -> str:
    """Build point-in-time query between each feature view table and the entity dataframe for teradata"""
    template = Environment(loader=BaseLoader()).from_string(source=query_template)

    final_output_feature_names = list(entity_df_columns)
    final_output_feature_names.extend(
        [
            (
                f'{fv["name"]}__{fv["field_mapping"].get(feature, feature)}'
                if full_feature_names
                else fv["field_mapping"].get(feature, feature)
            )
            for fv in feature_view_query_contexts
            for feature in fv["features"]
        ]
    )

    # Add additional fields to dict
    template_context = {
        "left_table_query_string": left_table_query_string,
        "entity_df_event_timestamp_col": entity_df_event_timestamp_col,
        "unique_entity_keys": set(
            [entity for fv in feature_view_query_contexts for entity in fv["entities"]]
        ),
        "featureviews": feature_view_query_contexts,
        "full_feature_names": full_feature_names,
        "final_output_feature_names": final_output_feature_names,
    }

    query = template.render(template_context)
    return query


def _upload_entity_df(
        config: RepoConfig, entity_df: Union[pd.DataFrame, str], table_name: str
):
    if isinstance(entity_df, pd.DataFrame):
        # If the entity_df is a pandas dataframe, upload it to Postgres
        df_to_teradata_table(config.offline_store, entity_df, table_name)
    elif isinstance(entity_df, str):
        with get_conn(config.offline_store).raw_connection().cursor() as cur:
            cur.execute(f"CREATE TABLE {table_name} AS ({entity_df}) with data")

    #     # If the entity_df is a string (SQL query), create a Postgres table out of it
    #         cur.execute(f"CREATE TABLE {table_name} AS ({entity_df})")
    else:
        raise InvalidEntityType(type(entity_df))


def _get_entity_schema(
        entity_df: Union[pd.DataFrame, str],
        config: RepoConfig,
) -> Dict[str, np.dtype]:
    if isinstance(entity_df, pd.DataFrame):
        return dict(zip(entity_df.columns, entity_df.dtypes))

    elif isinstance(entity_df, str):
        df_query = f"({entity_df}) AS sub"
        return get_query_schema(config.offline_store, df_query)
    else:
        raise InvalidEntityType(type(entity_df))


def get_query_schema(config: TeradataConfig, sql_query: str) -> Dict[str, np.dtype]:
    """
    We'll use the statement when we perform the query rather than copying data to a
    new table
    """
    with get_conn(config).connect() as conn:
        df = pd.read_sql(
            f"SELECT * FROM {sql_query}",
            conn,
        )
    return dict(zip(df.columns, df.dtypes))


# Copied from the Feast Redshift offline store implementation
# Note: Keep this in sync with sdk/python/feast/infra/offline_stores/redshift.py:
# MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN
# https://github.com/feast-dev/feast/blob/master/sdk/python/feast/infra/offline_stores/redshift.py

MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
/*
Compute a deterministic hash for the `left_table_query_string` that will be used throughout
all the logic as the field to GROUP BY the data
*/
WITH "entity_dataframe" AS (
    SELECT a.*,
        "{{entity_df_event_timestamp_col}}" AS "entity_timestamp"
        {% for featureview in featureviews %}
            {% if featureview.entities %}
            ,(
                {% for entity in featureview.entities %}
                    CAST("{{entity}}" AS VARCHAR(256)) ||
                {% endfor %}
                CAST("{{entity_df_event_timestamp_col}}" AS VARCHAR(256))
            ) AS "{{featureview.name}}__entity_row_unique_id"
            {% else %}
            ,CAST("{{entity_df_event_timestamp_col}}" AS VARCHAR(1000)) AS "{{featureview.name}}__entity_row_unique_id"
            {% endif %}
        {% endfor %}
    FROM "{{ left_table_query_string }}" a
),
{% for featureview in featureviews %}
"{{ featureview.name }}__entity_dataframe" AS (
    SELECT
       {{ featureview.entities | map('tojson') | join(', ')}}{% if featureview.entities %},{% else %}{% endif %}
        "entity_timestamp",
        "{{featureview.name}}__entity_row_unique_id"
    FROM "entity_dataframe"
    GROUP BY
        {{ featureview.entities | map('tojson') | join(', ')}}{% if featureview.entities %},{% else %}{% endif %}
        "entity_timestamp",
        "{{featureview.name}}__entity_row_unique_id"
),
/*
This query template performs the point-in-time correctness join for a single feature set table
to the provided entity table.
1. We first join the current feature_view to the entity dataframe that has been passed.
This JOIN has the following logic:
    - For each row of the entity dataframe, only keep the rows where the `timestamp_field`
    is less than the one provided in the entity dataframe
    - If there a TTL for the current feature_view, also keep the rows where the `timestamp_field`
    is higher the the one provided minus the TTL
    - For each row, Join on the entity key and retrieve the `entity_row_unique_id` that has been
    computed previously
The output of this CTE will contain all the necessary information and already filtered out most
of the data that is not relevant.
*/
"{{ featureview.name }}__subquery" AS (
    SELECT
        "{{ featureview.timestamp_field }}" as "event_timestamp",
        {{'"' ~ featureview.created_timestamp_column ~ '" as "created_timestamp",' if featureview.created_timestamp_column else '' }}
        {{featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% else %}{% endif %}
        {% for feature in featureview.features %}
            "{{ feature }}" as {% if full_feature_names %}"{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}"{% else %}"{{ featureview.field_mapping.get(feature, feature) }}"{% endif %}{% if loop.last %}{% else %}, {% endif %}
        {% endfor %}
    FROM {{ featureview.table_subquery }} as base
    WHERE "{{ featureview.timestamp_field }}" <= '{{ featureview.max_event_timestamp }}'
    {% if featureview.ttl == 0 %}{% else %}
    AND "{{ featureview.timestamp_field }}" >= '{{ featureview.min_event_timestamp }}'
    {% endif %}
),
"{{ featureview.name }}__base" AS (
    SELECT
        "subquery".*,
        "entity_dataframe"."entity_timestamp",
        "entity_dataframe"."{{featureview.name}}__entity_row_unique_id"
    FROM "{{ featureview.name }}__subquery" AS "subquery"
    INNER JOIN "{{ featureview.name }}__entity_dataframe" AS "entity_dataframe"
    ON 1=1
        AND "subquery"."event_timestamp" <= "entity_dataframe"."entity_timestamp"
        {% if featureview.ttl == 0 %}{% else %}
        AND subquery.event_timestamp >= entity_dataframe.entity_timestamp - {{ featureview.ttl }} * interval '0 00:00:01' day to second /* 
Uses TD specific function day to second to make sure any value for TTL works (up till 27 years converted to seconds). Takes an additional day parameter to convert the value to seconds
 */
        {% endif %}
        {% for entity in featureview.entities %}
        AND "subquery"."{{ entity }}" = "entity_dataframe"."{{ entity }}"
        {% endfor %}
),
/*
2. If the `created_timestamp_column` has been set, we need to
deduplicate the data first. This is done by calculating the
`MAX(created_at_timestamp)` for each event_timestamp.
We then join the data on the next CTE
*/
{% if featureview.created_timestamp_column %}
"{{ featureview.name }}__dedup" AS (
    SELECT
        "{{featureview.name}}__entity_row_unique_id",
        "event_timestamp",
        MAX("created_timestamp") AS "created_timestamp"
    FROM "{{ featureview.name }}__base"
    GROUP BY "{{featureview.name}}__entity_row_unique_id", "event_timestamp"
),
{% endif %}
/*
3. The data has been filtered during the first CTE "*__base"
Thus we only need to compute the latest timestamp of each feature.
*/
"{{ featureview.name }}__latest" AS (
    SELECT
        "event_timestamp",
        {% if featureview.created_timestamp_column %}"created_timestamp",{% endif %}
        "{{featureview.name}}__entity_row_unique_id"
    FROM
    (
        SELECT b.*,
            ROW_NUMBER() OVER(
                PARTITION BY b."{{featureview.name}}__entity_row_unique_id"
                ORDER BY b."event_timestamp" DESC{% if featureview.created_timestamp_column %},b."created_timestamp" DESC{% endif %}
            ) AS "row_number"
        FROM "{{ featureview.name }}__base" b
        {% if featureview.created_timestamp_column %}
            INNER JOIN "{{ featureview.name }}__dedup" bb
            ON b."{{featureview.name}}__entity_row_unique_id" = bb."{{featureview.name}}__entity_row_unique_id"
            AND b."event_timestamp" = bb."event_timestamp"
            AND b."created_timestamp" = bb."created_timestamp"
        {% endif %}
    ) as c
    WHERE "row_number" = 1
),
/*
4. Once we know the latest value of each feature for a given timestamp,
we can join again the data back to the original "base" dataset
*/
"{{ featureview.name }}__cleaned" AS (
    SELECT "base".*
    FROM "{{ featureview.name }}__base" AS "base"
    INNER JOIN "{{ featureview.name }}__latest"
    ON "base"."{{featureview.name}}__entity_row_unique_id" = "{{ featureview.name }}__latest"."{{featureview.name}}__entity_row_unique_id"
    AND "base"."event_timestamp" = "{{ featureview.name }}__latest"."event_timestamp"

    {% if featureview.created_timestamp_column %}
        AND
        "base"."created_timestamp" = "{{ featureview.name }}__latest"."created_timestamp"
    {% endif %}
){% if loop.last %}{% else %}, {% endif %}
{% endfor %}
/*
Joins the outputs of multiple time travel joins to a single table.
The entity_dataframe dataset being our source of truth here.
*/
SELECT "{{ final_output_feature_names | join('", "')}}"
FROM "entity_dataframe"
{% for featureview in featureviews %}
LEFT JOIN (
    SELECT
        "{{featureview.name}}__entity_row_unique_id"
        {% for feature in featureview.features %}
            ,{% if full_feature_names %}"{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}"{% else %}"{{ featureview.field_mapping.get(feature, feature) }}"{% endif %}
        {% endfor %}
    FROM "{{ featureview.name }}__cleaned"
) "{{ featureview.name }}__cleaned" ON "entity_dataframe"."{{featureview.name}}__entity_row_unique_id"="{{ featureview.name }}__cleaned"."{{featureview.name}}__entity_row_unique_id"
{% endfor %}
"""



from datetime import datetime
from typing import Sequence, List, Optional, Tuple, Dict, Callable, Any

import pytz
import itertools
from binascii import hexlify
from feast.usage import log_exceptions_and_usage
from feast import RepoConfig, FeatureView, Entity
from feast.infra.key_encoding_utils import serialize_entity_key
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from pydantic.typing import Literal
import pandas as pd
from feast.utils import to_naive_utc
from feast_teradata.teradata_utils import (
    get_conn,
    TeradataConfig
)
from teradataml import (
    copy_to_sql,
    VARBYTE,
    VARCHAR,
    TIMESTAMP,
    DataFrame
)

types_dict = {
    "entity_feature_key": VARBYTE(512),
    "entity_key": VARBYTE(512),
    "feature_name": VARCHAR(512),
    "value": VARBYTE(1024),
    "event_ts": TIMESTAMP,
    "created_ts": TIMESTAMP
}


class TeradataOnlineStoreConfig(TeradataConfig):
    type: Literal[
        "feast_teradata.online.teradata.TeradataOnlineStore"
    ] = "feast_teradata.online.teradata.TeradataOnlineStore"


class TeradataOnlineStore(OnlineStore):

    @log_exceptions_and_usage(online_store="teradata")
    def online_write_batch(
            self,
            config: RepoConfig,
            table: FeatureView,
            data: List[
                Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
            ],
            progress: Optional[Callable[[int], Any]],
    ) -> None:
        assert isinstance(config.online_store, TeradataOnlineStoreConfig)

        dfs = [None] * len(data)
        for i, (entity_key, values, timestamp, created_ts) in enumerate(data):
            df = pd.DataFrame(
                columns=[
                    "entity_feature_key",
                    "entity_key",
                    "feature_name",
                    "value",
                    "event_ts",
                    "created_ts",
                ],
                index=range(0, len(values)),
            )

            timestamp = to_naive_utc(timestamp)
            if created_ts is not None:
                created_ts = to_naive_utc(created_ts)

            for j, (feature_name, val) in enumerate(values.items()):
                df.loc[j, "entity_feature_key"] = serialize_entity_key(
                    entity_key,
                    entity_key_serialization_version=config.entity_key_serialization_version,
                ) + bytes(feature_name, encoding="utf-8")
                df.loc[j, "entity_key"] = serialize_entity_key(
                    entity_key,
                    entity_key_serialization_version=config.entity_key_serialization_version,
                )
                df.loc[j, "feature_name"] = feature_name
                df.loc[j, "value"] = val.SerializeToString()
                df.loc[j, "event_ts"] = timestamp
                df.loc[j, "created_ts"] = created_ts

            dfs[i] = df

        if dfs:
            agg_df = pd.concat(dfs)

            # This combines both the data upload plus the overwrite in the same transaction
            with get_conn(config.online_store).connect() as conn:
                copy_to_sql(df=agg_df,
                            table_name=f"{config.project}_{table.name}_t",
                            if_exists="replace",
                            types=types_dict,
                            primary_index="entity_feature_key")

                query = f"""
                        MERGE INTO {config.project}_{table.name} tar
                        USING {config.project}_{table.name}_t src
                           ON tar.entity_feature_key=src.entity_feature_key AND tar.entity_key = src.entity_key AND tar.feature_name = src.feature_name 
                        WHEN MATCHED THEN
                           UPDATE SET "value" = src."value", event_ts = src.event_ts, created_ts = src.created_ts
                        WHEN NOT MATCHED THEN
                               INSERT (entity_feature_key, entity_key, feature_name, "value", event_ts, created_ts) 
                                   VALUES (src.entity_feature_key, src.entity_key, src.feature_name, src."value", src.event_ts, src.created_ts)
                        """
                conn.execute(query)

            if progress:
                progress(len(data))

        return None

    @log_exceptions_and_usage(online_store="teradata")
    def online_read(
            self,
            config: RepoConfig,
            table: FeatureView,
            entity_keys: List[EntityKeyProto],
            requested_features: List[str],
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        assert isinstance(config.online_store, TeradataOnlineStoreConfig)

        result: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []

        entity_fetch_str = ",".join(
            [
                (
                        "TO_BYTES("
                        + hexlify(
                            serialize_entity_key(
                                combo[0],
                                entity_key_serialization_version=config.entity_key_serialization_version,
                            )
                            + bytes(combo[1], encoding="utf-8")
                        ).__str__()[1:]
                        + ",'base16')"
                )
                for combo in itertools.product(entity_keys, requested_features)
            ]
        )

        with get_conn(config.online_store).connect() as conn:
            query = f"""
                    SELECT
                        "entity_key", "feature_name", "value", "event_ts"
                    FROM
                        "{config.project}_{table.name}"
                    WHERE
                        "entity_feature_key" IN ({entity_fetch_str})
                """
            df = DataFrame.from_query(query).to_pandas()

        for entity_key in entity_keys:
            entity_key_bin = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            res = {}
            res_ts = None
            for index, row in df[df["entity_key"] == entity_key_bin].iterrows():
                val = ValueProto()
                val.ParseFromString(row["value"])
                res[row["feature_name"]] = val
                res_ts = row["event_ts"].to_pydatetime()

            if not res:
                result.append((None, None))
            else:
                result.append((res_ts, res))
        return result

    @log_exceptions_and_usage(online_store="teradata")
    def update(
            self,
            config: RepoConfig,
            tables_to_delete: Sequence[FeatureView],
            tables_to_keep: Sequence[FeatureView],
            entities_to_delete: Sequence[Entity],
            entities_to_keep: Sequence[Entity],
            partial: bool,
    ):
        assert isinstance(config.online_store, TeradataOnlineStoreConfig)

        with get_conn(config.online_store).connect() as conn:
            for table in tables_to_keep:
                query = f"""
                        CREATE TABLE {config.project}_{table.name} (
                            "entity_feature_key" VARBYTE(512),
                            "entity_key" VARBYTE(512),
                            "feature_name" VARCHAR(512),
                            "value" VARBYTE(1024),
                            "event_ts" TIMESTAMP,
                            "created_ts" TIMESTAMP
                        )
                    """
                conn.execute(query)

            for table in tables_to_delete:
                query = f"""DROP TABLE {config.project}_{table.name}"""
                conn.execute(query)

    def teardown(
            self,
            config: RepoConfig,
            tables: Sequence[FeatureView],
            entities: Sequence[Entity],
    ):
        assert isinstance(config.online_store, TeradataOnlineStoreConfig)

        with get_conn(config.online_store).connect() as conn:
            for table in tables:
                query = f"""DROP TABLE {config.project}_{table.name}"""
                conn.execute(query)


def _to_naive_utc(ts: datetime):
    if ts.tzinfo is None:
        return ts
    else:
        return ts.astimezone(pytz.utc).replace(tzinfo=None)


def _table_id(project: str, table: FeatureView) -> str:
    return f"{project}_{table.name}"



import json
from typing import Callable, Dict, Iterable, Optional, Tuple
from typeguard import typechecked

from feast.data_source import DataSource
from feast.errors import DataSourceNoNameException
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.value_type import ValueType
import pandas as pd
import numpy as np

from feast_teradata.teradata_utils import (
    get_conn,
    TeradataConfig
)
from teradataml import DataFrame


def td_type_to_feast_value_type(type_str: str) -> ValueType:
    type_map: Dict[str, ValueType] = {
        "<class 'byte'>": ValueType.BYTES,
        "<class 'str'>": ValueType.STRING,
        "<class 'int'>": ValueType.INT64,
        "<class 'float'>": ValueType.FLOAT,
        "<class 'datetime.datetime'>": ValueType.UNIX_TIMESTAMP,
    }
    value = (
        type_map[f"""{type_str}"""]
        if f"""{type_str}""" in type_map
        else ValueType.UNKNOWN
    )
    if value == ValueType.UNKNOWN:
        print("unknown type:", type_str)
    return value
@typechecked
class TeradataSource(DataSource):
    def __init__(
            self,
            name: Optional[str] = None,
            query: Optional[str] = None,
            table: Optional[str] = None,
            database: Optional[str] = None,
            timestamp_field: Optional[str] = "",
            created_timestamp_column: Optional[str] = "",
            field_mapping: Optional[Dict[str, str]] = None,
            description: Optional[str] = "",
            tags: Optional[Dict[str, str]] = None,
            owner: Optional[str] = "",
    ):
        self._teradata_options = TeradataOptions(name=name,
                                                 query=query,
                                                 database=database,
                                                 table=table)

        # If no name, use the table as the default name.
        if name is None and table is None:
            raise DataSourceNoNameException()
        name = name or table
        assert name

        super().__init__(
            name=name,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            description=description,
            tags=tags,
            owner=owner,
        )

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, TeradataSource):
            raise TypeError(
                "Comparisons should only involve TeradataSource class objects."
            )

        return (
                super().__eq__(other)
                and self._teradata_options._query == other._teradata_options._query
                and self.timestamp_field == other.timestamp_field
                and self.created_timestamp_column == other.created_timestamp_column
                and self.field_mapping == other.field_mapping
        )

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        assert data_source.HasField("custom_options")

        teradata_options = json.loads(data_source.custom_options.configuration)

        return TeradataSource(
            name=teradata_options["name"],
            query=teradata_options["query"],
            table=teradata_options["table"],
            database=teradata_options["database"],
            field_mapping=dict(data_source.field_mapping),
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
        )

    def get_table_query_string(self) -> str:
        """Returns a string that can directly be used to reference this table in SQL."""
        if self._teradata_options._table:
            return f"{self._teradata_options._table}"
        else:
            return f"({self._teradata_options._query})"

    def to_proto(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.CUSTOM_SOURCE,
            data_source_class_type="feast_teradata.offline.teradata_source.TeradataSource",
            field_mapping=self.field_mapping,
            custom_options=self._teradata_options.to_proto(),
            description=self.description,
            tags=self.tags,
            owner=self.owner,
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    def validate(self, config: RepoConfig):
        pass


    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return td_type_to_feast_value_type

    def get_table_column_names_and_types(
            self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        with get_conn(config.offline_store).raw_connection().cursor() as cur:
            # df = pd.read_sql(f"SELECT * FROM {self.get_table_query_string()} sample 1", conn)
            # column_names = df.columns
            # types = df.dtypes
            # return list(zip(column_names, types))
            cur.execute(
                f"SELECT * FROM {self.get_table_query_string()} sample 1"
            )
            name_desc_ret = [(item[0], item[1]) for item in cur.description]

            return name_desc_ret
class TeradataOptions:
    def __init__(
            self,
            name: Optional[str],
            query: Optional[str],
            table: Optional[str],
            database: Optional[str]
    ):
        self._name = name or ""
        self._query = query or ""
        self._table = table or ""
        self._database = database or ""

    @classmethod
    def from_proto(cls, teradata_options_proto: DataSourceProto.CustomSourceOptions):
        config = json.loads(teradata_options_proto.configuration.decode("utf8"))
        teradata_options = cls(
            name=config["name"],
            query=config["query"],
            table=config["table"],
            database=config["database"]
        )

        return teradata_options

    def to_proto(self) -> DataSourceProto.CustomSourceOptions:
        teradata_options_proto = DataSourceProto.CustomSourceOptions(
            configuration=json.dumps(
                {"name": self._name,
                 "query": self._query,
                 "table": self._table,
                 "database": self._database}
            ).encode()
        )
        return teradata_options_proto


class SavedDatasetTeradataStorage(SavedDatasetStorage):
    _proto_attr_name = "custom_storage"

    teradata_options: TeradataOptions

    def __init__(self, table_ref: str):
        self.teradata_options = TeradataOptions(
            table=table_ref,
            name=None,
            query=None,
            database=None
        )

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:
        return SavedDatasetTeradataStorage(
            table_ref=TeradataOptions.from_proto(storage_proto.custom_storage)._table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(custom_storage=self.teradata_options.to_proto())

    def to_data_source(self) -> DataSource:
        return TeradataSource(table=self.teradata_options._table)


def df_to_teradata_table(config: TeradataConfig, df: pd.DataFrame, table_name: str) -> Dict[str, np.dtype]:
    """
    Create a table for the data frame, insert all the values, and return the table schema
    """

    with get_conn(config).connect() as conn:
        col_type_dict = dict(zip(df.columns, df.dtypes))
        df.to_sql(name=table_name, con=conn, if_exists='replace', index=False)

        return col_type_dict



from setuptools import find_packages, setup
from feast_teradata import __version__

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md") as f:
    long_description = f.read()

setup(
    name="feast-teradata",
    version=__version__,
    author="Teradata Corporation",
    author_email="developers@teradata.com",
    url="https://github.com/Teradata/feast-teradata",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8,<3.11",
    packages=find_packages(exclude=('tests',)),
    package_data={
        'feast_teradata.cli': [
            'template/teradata/**',
            'template/teradata/**/**',
        ],
    },
    install_requires=required,
    tests_require=['pytest==6.2.4'],
    license_files=['LICENSE', 'LICENSE-3RD-PARTY.txt'],
    entry_points={'console_scripts': [
        'feast-td=feast_teradata.cli.cli:cli',
    ]},
)



import sys
import click
import re

from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from click.exceptions import BadParameter
from feast.file_utils import replace_str_in_file


def is_valid_name(name: str) -> bool:
    """A name should be alphanumeric values and underscores but not start with an underscore"""
    return not name.startswith("_") and re.compile(r"\W+").search(name) is None


@click.group()
def cli():
    pass


@cli.command()
def init_repo():
    import os
    import importlib.util
    from distutils.dir_util import copy_tree
    from pathlib import Path
    from colorama import Fore, Style

    # check if running in interactive mode or via CI
    is_interactive = os.environ.get("CI_RUNNING", "false").lower() == "false"

    template = "teradata"

    if is_interactive:
        repo_name = click.prompt("Repository Name:")
    else:
        repo_name = os.environ.get("CI_FEAST_REPO_NAME")

    if not is_valid_name(repo_name):
        raise BadParameter(
            message="Name should be alphanumeric values and underscores but not start with an underscore",
            param_hint="PROJECT_DIRECTORY",
        )
    repo_path = Path(os.path.join(Path.cwd(), repo_name))
    repo_path.mkdir(exist_ok=True)
    repo_config_path = repo_path / "feature_store.yaml"

    if repo_config_path.exists():
        new_directory = os.path.relpath(repo_path, os.getcwd())

        print(
            f"The directory {Style.BRIGHT + Fore.GREEN}{new_directory}{Style.RESET_ALL} contains an existing feature "
            f"store repository that may cause a conflict"
        )
        print()
        sys.exit(1)

    # Copy template directory
    template_path = str(Path(Path(__file__).parent / "template" / template).absolute())
    copy_tree(template_path, str(repo_path))

    # Seed the repository
    bootstrap_path = repo_path / "bootstrap.py"
    if not os.path.exists(bootstrap_path):
        raise Exception(f"boostrap.py not found at {bootstrap_path}")

    spec = importlib.util.spec_from_file_location("bootstrap", str(bootstrap_path))
    assert isinstance(spec, ModuleSpec)
    bootstrap = importlib.util.module_from_spec(spec)
    assert isinstance(spec.loader, Loader)
    spec.loader.exec_module(bootstrap)
    bootstrap.bootstrap()  # type: ignore
    os.remove(bootstrap_path)

    # template the feature_store.yaml file
    feature_store_yaml_path = repo_path / "feature_repo" / "feature_store.yaml"
    replace_str_in_file(
        feature_store_yaml_path, "project: my_project", f"project: {repo_name}"
    )

    # Remove the __pycache__ folder if it exists
    import shutil
    shutil.rmtree(repo_path / "__pycache__", ignore_errors=True)

    click.echo()
    click.echo(
        f"Creating a new Feast repository in {Style.BRIGHT + Fore.GREEN}{repo_path}{Style.RESET_ALL}."
    )
    click.echo()


if __name__ == '__main__':
    cli()

